@@ -20,4 +20,4 @@ finally:
 from .score import Score
 from .annotations import Annotations
 from .parse import Parse
-from .utils import COLORS, load_tsv
+from .utils import *
