@@ -2,7 +2,8 @@ import sys, re
 
 import pandas as pd
 
-from .utils import DCML_REGEX, DCML_DOUBLE_REGEX, decode_harmonies, is_any_row_equal, html2format, load_tsv, map_dict, name2format, resolve_dir, rgb2format, update_cfg
+from .utils import DCML_REGEX, DCML_DOUBLE_REGEX, decode_harmonies, is_any_row_equal, html2format, load_tsv, \
+    map_dict, name2format, resolve_dir, rgb2format, column_order, update_cfg
 from .logger import LoggedClass
 from .expand_dcml import expand_labels
 
@@ -25,6 +26,11 @@ class Annotations(LoggedClass):
         cols : :obj:`dict`, optional
             If one or several column names differ, pass a {NAME -> ACTUAL_NAME} dictionary, where NAME can be
             {'mc', 'mn', 'mc_onset', 'label', 'staff', 'voice', 'volta'}
+        cols : :obj:`dict`
+            If your columns don't have standard names, pass a {NAME -> ACTUAL_NAME} dictionary.
+            Required columns: label, mc, mc_onset, staff, voice
+            Additional columns: label_type, root, base, leftParen, rightParen, offset_x, offset_y, nashville, color_name,
+            color_html, color_r, color_g, color_b, color_a
         index_col
         sep
         mscx_obj
@@ -43,25 +49,41 @@ class Annotations(LoggedClass):
         self.changed = False
         self.read_only = read_only
         self.mscx_obj = mscx_obj
+        self.volta_structure = None if mscx_obj is None else mscx_obj.volta_structure
+
         columns = self.main_cols + self.additional_cols
         self.cols = {c: c for c in columns}
         self.cols.update(update_cfg(cols, self.cols.keys(), logger=self.logger))
 
+
         if df is not None:
             self.df = df.copy()
         else:
             assert tsv_path is not None, "Name a TSV file to be loaded."
             self.df = load_tsv(tsv_path, index_col=index_col, sep=sep, **kwargs)
-        for col in ['label']:
-            assert self.cols[col] in self.df.columns, f"""The DataFrame has no column named '{self.cols[col]}'. Pass the column name as col={{'{col}'=col_name}}.
-Present column names are:\n{self.df.columns.to_list()}."""
+        sorting_cols = ['mc', 'mn', 'mc_onset', 'staff']
+        sorting_cols = [self.cols[c] if c in self.cols else c for c in sorting_cols]
+        sorting_cols = [c for c in sorting_cols if c in self.df.columns]
+        self.df.sort_values(sorting_cols, inplace=True)
+#         for col in ['label']:
+#             assert self.cols[col] in self.df.columns, f"""The DataFrame has no column named '{self.cols[col]}'. Pass the column name as col={{'{col}'=col_name}}.
+# Present column names are:\n{self.df.columns.to_list()}."""
         # if 'offset' in self.df.columns:
         #     self.df.drop(columns='offset', inplace=True)
         #self.cols = {k: v for k, v in self.cols.items() if k in self.main_cols + ['label_type'] or v in df.columns}
         self.infer_types()
 
+    def add_initial_dots(self):
+        if self.read_only:
+            self.logger.warning(f"Cannot change labels attached to a score. Detach them first.")
+            return
+        label_col = self.cols['label']
+        notes = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'}
+        add_dots = lambda s: '.' + s if s[0].lower() in notes else s
+        self.df[label_col] = self.df[label_col].map(add_dots)
 
-    def prepare_for_attaching(self, staff=None, voice=None, check_for_clashes=True):
+
+    def prepare_for_attaching(self, staff=None, voice=None, label_type=None, check_for_clashes=True):
         if self.mscx_obj is None:
             self.logger.warning(f"Annotations object not aware to which MSCX object it is attached.")
             return pd.DataFrame()
@@ -89,11 +111,14 @@ Possible values are {{1, 2, 3, 4}}.""")
         if voice is not None:
             df[voice_col] = voice
         if voice_col in cols and df[voice_col].isna().any():
-            self.logger.warning(f"The following labels don't ahve voice information: {df[df.voice.isna()]}")
+            self.logger.warning(f"The following labels don't have voice information: {df[df.voice.isna()]}")
             error = True
         if error:
             return pd.DataFrame()
 
+        if label_type is not None:
+            df[self.cols['label_type']] = label_type
+
         if self.cols['mc'] not in cols:
             mn_col = self.cols['mn'] if 'mn' in self.cols else 'mn'
             if mn_col not in cols:
@@ -162,7 +187,10 @@ Possible values are {{1, 2, 3, 4}}.""")
             elif self.cols['color_r'] in df.columns:
                 df[color_name] = rgb2format(df, 'name')
             df[color_name] = df[color_name].fillna('default')
-            layers += [color_name]
+            layers.append(color_name)
+        else:
+            df['color_name'] = 'default'
+            layers.append('color_name')
         type2name = map_dict({
             0: '0 (Plain Text)',
             1: '1 (Nashville)',
@@ -177,7 +205,7 @@ Possible values are {{1, 2, 3, 4}}.""")
         n, layers = self.annotation_layers
         return f"{n} labels:\n{layers.to_string()}"
 
-    def get_labels(self, staff=None, voice=None, label_type=None, positioning=True, decode=False, drop=False, warnings=True, column_name=None, color_format='html'):
+    def get_labels(self, staff=None, voice=None, label_type=None, positioning=True, decode=False, drop=False, warnings=True, column_name=None, color_format=None):
         """ Returns a DataFrame of annotation labels.
 
         Parameters
@@ -234,30 +262,43 @@ Possible values are {{1, 2, 3, 4}}.""")
             res = res.rename(columns={label_col: column_name})
         color_cols = ['color_html', 'color_r', 'color_g', 'color_b', 'color_a', 'color_name']
         rgb_cols = ['color_r', 'color_g', 'color_b']
-        if color_format is not None and any(True for c in color_cols if c in res):
-            if color_format == 'html' and 'color_html' not in res.columns:
-                if 'color_name' in res.columns:
-                    html = name2format(res, 'html')
-                elif 'color_r' in res.columns:
-                    if any(True for c in rgb_cols if c not in res.columns):
-                        logger.warning(f"The following columns are missing: {list(c for c in rgb_cols if c not in res.columns)}")
-                    else:
-                        html = rgb2format(res, 'html')
-                res['color_html'] = html
-            elif color_format == 'name' and 'color_name' not in res.columns:
-                if 'color_html' in res.columns:
-                    name = html2format(res, 'name')
-                elif 'color_r' in res.columns:
-                    if any(True for c in rgb_cols if c not in res.columns):
-                        logger.warning(f"The following columns are required")
-                    else:
-                        name = rgb2format(res, 'name')
-                res['color_name'] = name
+        present_cols = [c for c in color_cols if c in res.columns]
+        if color_format is not None and len(present_cols) > 0:
+            res['color'] = pd.NA
+            has_html = 'color_html' in res.columns
+            has_name = 'color_name' in res.columns
+            has_rgb  = all(col in res.columns for col in rgb_cols)
+            has_rgba = has_rgb and 'color_a' in res.columns
 
+            def tuple_or_na(row):
+                if row.isna().all():
+                    return pd.NA
+                return tuple(row)
+
+            if color_format == 'html' and has_html:
+                res.color = res.color_html
+            elif color_format == 'name' and has_name:
+                res.color = res.color_name
+            elif color_format == 'rgb' and has_rgb:
+                res.color = res[rgb_cols].apply(tuple_or_na, axis=1)
+            elif color_format == 'rgba' and has_rgba:
+                res.color = res[rgb_cols + ['color_a']].apply(tuple_or_na, axis=1)
+            elif has_html:
+                res.color = html2format(res, color_format)
+            elif has_name:
+                res.color = name2format(res, color_format)
+            elif has_rgb:
+                res.color = rgb2format(res, color_format)
+            else:
+                logger.warning(f"Color format '{color_format}' could not be computed from columns {present_cols}.")
+            res.drop(columns=present_cols, inplace=True)
+
+        if self.mscx_obj is not None:
+            res = column_order(self.mscx_obj.parsed.add_standard_cols(res))
         return res
 
 
-    def expand_dcml(self, drop_others=True, warn_about_others=True, **kwargs):
+    def expand_dcml(self, drop_others=True, warn_about_others=True, drop_empty_cols=False, **kwargs):
         """ Expands all labels where the label_type has been inferred as 'dcml' and stores the DataFrame in self._expanded.
 
         Parameters
@@ -267,6 +308,8 @@ Possible values are {{1, 2, 3, 4}}.""")
         warn_about_others : :obj:`bool`, optional
             Set to False to suppress warnings about labels that have not label_type 'dcml'.
             Is automatically set to False if ``drop_others`` is set to False.
+        drop_empty_cols : :obj:`bool`, optional
+            Return without unused columns
         kwargs
             Additional arguments are passed to :py:meth:`.get_labels` to define the original representation.
 
@@ -289,7 +332,7 @@ Possible values are {{1, 2, 3, 4}}.""")
             self.logger.warning(f"Score contains {(~sel).sum()} labels that don't (and {sel.sum()} that do) match the DCML standard:\n{decode_harmonies(df[~sel], keep_type=True)[['mc', 'mn', 'label', 'label_type']].to_string()}")
         df = df[sel]
         try:
-            exp = expand_labels(df, column='label', regex=DCML_REGEX, chord_tones=True, logger=self.logger)
+            exp = expand_labels(df, column='label', regex=DCML_REGEX, volta_structure=self.volta_structure, chord_tones=True, logger=self.logger)
             if drop_others:
                 self._expanded = exp
             else:
@@ -301,6 +344,8 @@ Possible values are {{1, 2, 3, 4}}.""")
         except:
             self.logger.error(f"Expanding labels failed with the following error:\n{sys.exc_info()[1]}")
 
+        if drop_empty_cols:
+            return self._expanded.dropna(axis=1, how='all')
         return self._expanded
 
 
@@ -320,26 +365,37 @@ Possible values are {{1, 2, 3, 4}}.""")
 
 
     def infer_types(self, regex_dict=None):
+        recognized = [0, 1, 2, 3, '0', '1', '2', '3']
         if regex_dict is not None:
             self.regex_dict = regex_dict
         if 'label_type' in self.df.columns:
             self.df.label_type.fillna(0, inplace=True)
-            self.df.loc[~self.df.label_type.isin([0, 1, 2, 3, '0', '1', '2', '3']), 'label_type'] = 0
+            self.df.loc[~self.df.label_type.isin(recognized), 'label_type'] = 0
         else:
             self.df['label_type'] = pd.Series(0, index=self.df.index, dtype='object')
         if 'nashville' in self.df.columns:
             self.df.loc[self.df.nashville.notna(), 'label_type'] = 2
-        if 'root' in self.df.columns:
-            self.df.loc[self.df.root.notna(), 'label_type'] = 3
+        if 'absolute_root' in self.df.columns:
+            self.df.loc[self.df.absolute_root.notna(), 'label_type'] = 3
         if len(self.regex_dict) > 0:
             decoded = decode_harmonies(self.df, label_col=self.cols['label'], return_series=True)
             for name, regex in self.regex_dict.items():
-                sel = self.df.label_type.isin((0, 1, 2, 3))
+                sel = self.df.label_type.isin(recognized)
                 #mtch = self.df.loc[sel, self.cols['label']].str.match(regex)
                 mtch = decoded[sel].str.match(regex)
                 self.df.loc[sel & mtch, 'label_type'] = self.df.loc[sel & mtch, 'label_type'].astype(str) + f" ({name})"
 
 
+    def remove_initial_dots(self):
+        if self.read_only:
+            self.logger.warning(f"Cannot change labels attached to a score. Detach them first.")
+            return
+        label_col = self.cols['label']
+        rem_dots = lambda s: s[1:] if s[0] == '.' else s
+        no_nan = self.df[label_col].notna()
+        self.df.loc[no_nan, label_col] = self.df.loc[no_nan, label_col].map(rem_dots)
+
+
     def store_tsv(self, tsv_path, staff=None, voice=None, label_type=None, positioning=True, decode=False, sep='\t', index=False, **kwargs):
         df = self.get_labels(staff=staff, voice=voice, label_type=label_type, positioning=positioning, decode=decode)
         if decode and 'label_type' in df.columns:
